// .NAME vtkEdgeFaceCellIterator - iterator used to traverse cells
// .SECTION Description
// This class (and subclasses) are used to iterate over cells. Use it
// only in conjunction with vtkEdgeFaceDataSet (i.e., the adaptor framework).
//
// Typical use is:
// <pre>
// vtkEdgeFaceDataSet *dataset;
// vtkEdgeFaceCellIterator *it = dataset->NewCellIterator(2);
// for (it->Begin(); !it->IsAtEnd(); it->Next());
//   {
//   spec=it->GetCell();
//   }
// </pre>

#ifndef __vtkEdgeFaceCellIterator_h
#define __vtkEdgeFaceCellIterator_h

#include "vtkGenericCellIterator.h"

class vtkGenericAdaptorCell;
class vtkEdgeFaceDataSet;

class VTK_FILTERING_EXPORT vtkEdgeFaceCellIterator : public vtkGenericCellIterator
{
public:
  // Description:
  // Standard VTK construction and type macros.
  static vtkEdgeFaceCellIterator* New();
  vtkTypeRevisionMacro(vtkEdgeFaceCellIterator,vtkGenericCellIterator);
  void PrintSelf( ostream& os, vtkIndent indent );

  //BTX
  // Description:
  // Modes of iteration.
  enum IterationMode
    {
    MESH,     //!< Iterate over all cells in the mesh
    BOUNDARY  //!< Only iterate over cells on the mesh boundary
    };
  //ETX

  // Description:
  // Move iterator to first position if any (loop initialization).
  virtual void Begin();

  // Description:
  // Is the iterator at the end of traversal?
  virtual int IsAtEnd();
  
  // Description:
  // Create an empty cell. The user is responsible for deleting it.
  // \post result_exists: result!=0
  virtual vtkGenericAdaptorCell* NewCell();
  
  // Description:
  // Get the cell at current position. The cell should be instantiated
  // with the NewCell() method.
  // \pre not_at_end: !IsAtEnd()
  // \pre c_exists: c!=0
  // THREAD SAFE
  virtual void GetCell( vtkGenericAdaptorCell* c );
  
  // Description:
  // Get the cell at the current traversal position.
  // NOT THREAD SAFE
  // \pre not_at_end: !IsAtEnd()
  // \post result_exits: result!=0
  virtual vtkGenericAdaptorCell* GetCell();
  
  // Description:
  // Move the iterator to the next position in the list.
  // \pre not_at_end: !IsAtEnd()
  virtual void Next();

protected:
  vtkEdgeFaceCellIterator();
  virtual ~vtkEdgeFaceCellIterator();

  // Description:
  // Set/get the data set which owns the cell being iterated.
  // These methods should only be used by vtkEdgeFaceDataSet.
  virtual void SetDataSet( vtkEdgeFaceDataSet* dataSet );
  vtkEdgeFaceDataSet* GetDataSet() { return this->DataSet; }

  // Description:
  // Set/get the mode by which the iterator advances: does it
  // iterate over all cells or just boundaries of the mesh?
  // These methods should only be used by vtkEdgeFaceDataSet.
  vtkSetMacro(Mode,int);
  vtkGetMacro(Mode,int);
  
  //BTX 
  friend class vtkEdgeFaceDataSet;
  vtkEdgeFaceDataSet* DataSet;
  int Mode;
  //ETX 

private:
  vtkEdgeFaceCellIterator( const vtkEdgeFaceCellIterator& ); // Not implemented.
  void operator = ( const vtkEdgeFaceCellIterator& ); // Not implemented.
};

#endif
