// .NAME vtkEdgeFacePointIterator - iterator used to traverse points
// .SECTION Description
// This class (and subclasses) are used to iterate over points. Use it
// only in conjunction with vtkEdgeFaceDataSet (i.e., the adaptor framework).
//
// Typical use is:
// <pre>
// vtkEdgeFaceDataSet *dataset;
// vtkEdgeFacePointIterator *it = dataset->NewPointIterator();
// for (it->Begin(); !it->IsAtEnd(); it->Next());
//   {
//   x=it->GetPosition();
//   }
// </pre>


#ifndef __vtkEdgeFacePointIterator_h
#define __vtkEdgeFacePointIterator_h

#include "vtkGenericPointIterator.h"

class vtkEdgeFaceDataSet;

class VTK_FILTERING_EXPORT vtkEdgeFacePointIterator : public vtkGenericPointIterator
{
public:
  // Description:
  // Standard VTK construction and type macros.
  static vtkEdgeFacePointIterator* New();
  vtkTypeRevisionMacro(vtkEdgeFacePointIterator,vtkGenericPointIterator);
  void PrintSelf( ostream& os, vtkIndent indent );

  // Description:
  // Move iterator to first position if any (loop initialization).
  virtual void Begin();

  // Description:
  // Is the iterator at the end of traversal?
  virtual int IsAtEnd();

  // Description:
  // Move the iterator to the next position in the list.
  // \pre not_off: !IsAtEnd()
  virtual void Next();
 
  // Description:
  // Get the coordinates of the point at the current iterator position.
  // \pre not_off: !IsAtEnd()
  // \post result_exists: result!=0
  virtual double* GetPosition();

  // Description:
  // Get the coordinates of the point at the current iterator position.
  // \pre not_off: !IsAtEnd()
  // \pre x_exists: x!=0
  virtual void GetPosition( double x[3] );
  
  // Description:
  // Return the unique identifier for the point, could be non-contiguous.
  // \pre not_off: !IsAtEnd()
  virtual vtkIdType GetId();
  
protected:
  // Description:
  // Destructor.
  vtkEdgeFacePointIterator();
  virtual ~vtkEdgeFacePointIterator();
  
  virtual void SetDataSet( vtkEdgeFaceDataSet* dataSet );
  vtkEdgeFaceDataSet* GetDataSet() { return this->DataSet; }
 
  //BTX 
  friend class vtkEdgeFaceDataSet;
  vtkEdgeFaceDataSet* DataSet;
  //ETX

private:
  vtkEdgeFacePointIterator( const vtkEdgeFacePointIterator& ); // Not implemented.
  void operator = ( const vtkEdgeFacePointIterator& ); // Not implemented.
};

#endif
