// .NAME vtkEdgeFaceAttribute - abstract class defined API for attribute data
// .SECTION Description
// vtkEdgeFaceAttribute is an abstract class that defines an API for attribute
// data. Attribute data is data associated with the topology or geometry of
// a dataset (i.e., points, cells, etc.). vtkEdgeFaceAttribute is part of the
// adaptor framework (see GenericFiltering/README.html).
//
// vtkEdgeFaceAttribute provides a more general interface to attribute data
// than its counterpart vtkDataArray (which assumes a linear, contiguous
// array). It adopts an iterator interface, and allows attributes to be
// associated with points, edges, faces, or edges.

#ifndef __vtkEdgeFaceAttribute_h
#define __vtkEdgeFaceAttribute_h

#include "vtkGenericAttribute.h"

class vtkDoubleArray;
class vtkEdgeFaceDataSet;

class VTK_FILTERING_EXPORT vtkEdgeFaceAttribute : public vtkGenericAttribute
{
public:
  static vtkEdgeFaceAttribute* New();
  vtkTypeRevisionMacro(vtkEdgeFaceAttribute,vtkGenericAttribute );
  virtual void PrintSelf( ostream& os, vtkIndent indent );

  //BTX
  // Description:
  // Values that the DeRhamBasis may take on.
  // Only EDGE and FACE are currently supported.
  enum DeRhamBasisValues
    {
    CELL,
    FACE,
    EDGE,
    NODE
    };
  //ETX

  // Description:
  // Name of the attribute. (e.g. "velocity")
  // \post result_may_not_exist: result!=0 || result==0
  virtual const char* GetName();

  // Description:
  // Dimension of the attribute. (1 for scalar, 3 for velocity)
  // \post positive_result: result>=0
  // \post GetType()==VTK_SCALARS implies result==1
  // \post (GetType()==VTK_VECTORS||(GetType()==VTK_NORMALS)||(GetType()==VTK_TCOORDS) implies result==3
  // \post GetType()==VTK_TENSORS implies result==6
  virtual int GetNumberOfComponents();

  // Description:
  // Is the attribute centered either on points, cells or boundaries?
  // \post valid_result: (result==vtkPointCentered)||(result==vtkCellCentered)
  virtual int GetCentering();
  
  // Description:
  // Type of the attribute: scalar, vector, normal, texture coordinate, tensor
  // \post valid_result: (result==vtkDataSetAttributes::SCALARS)
  //                   ||(result==vtkDataSetAttributes::VECTORS)
  //                   ||(result==vtkDataSetAttributes::NORMALS)
  //                   ||(result==vtkDataSetAttributes::TCOORDS)
  //                   ||(result==vtkDataSetAttributes::TENSORS)
  virtual int GetType();
  
  // Description:
  // Type of the components of the attribute: int, float, double
  // \post valid_result: (result==VTK_BIT)           ||(result==VTK_CHAR)
  //                   ||(result==VTK_UNSIGNED_CHAR) ||(result==VTK_SHORT)
  //                   ||(result==VTK_UNSIGNED_SHORT)||(result==VTK_INT)
  //                   ||(result==VTK_UNSIGNED_INT)  ||(result==VTK_LONG)
  //                   ||(result==VTK_UNSIGNED_LONG) ||(result==VTK_FLOAT)
  //                   ||(result==VTK_DOUBLE)        ||(result==VTK_ID_TYPE)
  virtual int GetComponentType();

  // Description:
  // Number of tuples.
  // \post valid_result: result>=0
  virtual vtkIdType GetSize();

  // Description:
  // Size in kilobytes taken by the attribute.
  virtual unsigned long GetActualMemorySize();

  // Description:
  // Range of the attribute component `component'. If `component'==-1, it
  // returns the range of the magnitude (euclidean norm).
  // It returns double, even if GetType()==VTK_INT.
  // NOT THREAD SAFE
  // \pre valid_component: (component>=-1)&&(component<GetNumberOfComponents())
  // \post result_exists: result!=0
  virtual double* GetRange( int component = 0 );
  
  // Description:
  // Range of the attribute component `component'. If `component'==-1, it
  // returns the range of the magnitude (euclidean norm).
  // THREAD SAFE
  // \pre valid_component: (component>=-1)&&(component<GetNumberOfComponents())
  virtual void GetRange( int component, double range[2] );
  
  // Description:
  // Return the maximum euclidean norm for the tuples.
  // \post positive_result: result>=0
  virtual double GetMaxNorm();
  
  // Description:
  // Attribute at all points of cell `c'.
  // \pre c_exists: c!=0
  // \pre c_valid: !c->IsAtEnd()
  // \post result_exists: result!=0
  // \post valid_result: sizeof(result)==GetNumberOfComponents()*c->GetCell()->GetNumberOfPoints()
  virtual double* GetTuple( vtkGenericAdaptorCell* c );
  
  // Description:
  // Put attribute at all points of cell `c' in `tuple'.
  // \pre c_exists: c!=0
  // \pre c_valid: !c->IsAtEnd()
  // \pre tuple_exists: tuple!=0
  // \pre valid_tuple: sizeof(tuple)>=GetNumberOfComponents()*c->GetCell()->GetNumberOfPoints()
  virtual void GetTuple( vtkGenericAdaptorCell* c, double* tuple );
  
  // Description:
  // Attribute at all points of cell `c'.
  // \pre c_exists: c!=0
  // \pre c_valid: !c->IsAtEnd()
  // \post result_exists: result!=0
  // \post valid_result: sizeof(result)==GetNumberOfComponents()*c->GetCell()->GetNumberOfPoints()
  virtual double* GetTuple( vtkGenericCellIterator* c );

  // Description:
  // Put attribute at all points of cell `c' in `tuple'.
  // \pre c_exists: c!=0
  // \pre c_valid: !c->IsAtEnd()
  // \pre tuple_exists: tuple!=0
  // \pre valid_tuple: sizeof(tuple)>=GetNumberOfComponents()*c->GetCell()->GetNumberOfPoints()
  virtual void GetTuple( vtkGenericCellIterator* c, double* tuple );

  // Description:
  // Value of the attribute at position `p'.
  // \pre p_exists: p!=0
  // \pre p_valid: !p->IsAtEnd()
  // \post result_exists: result!=0
  // \post valid_result_size: sizeof(result)==GetNumberOfComponents()
  virtual double* GetTuple( vtkGenericPointIterator* p );
  
  // Description:
  // Put the value of the attribute at position `p' into `tuple'.
  // \pre p_exists: p!=0
  // \pre p_valid: !p->IsAtEnd()
  // \pre tuple_exists: tuple!=0
  // \pre valid_tuple_size: sizeof(tuple)>=GetNumberOfComponents()
  virtual void GetTuple( vtkGenericPointIterator* p, double* tuple );
  
  // Description:
  // Put component `i' of the attribute at all points of cell `c' in `values'.
  // \pre valid_component: (i>=0) && (i<GetNumberOfComponents())
  // \pre c_exists: c!=0
  // \pre c_valid: !c->IsAtEnd()
  // \pre values_exist: values!=0
  // \pre valid_values: sizeof(values)>=c->GetCell()->GetNumberOfPoints()
  virtual void GetComponent( int i, vtkGenericCellIterator* c, double* values );

  // Description:
  // Value of the component `i' of the attribute at position `p'.
  // \pre valid_component: (i>=0) && (i<GetNumberOfComponents())
  // \pre p_exists: p!=0
  // \pre p_valid: !p->IsAtEnd()
  virtual double GetComponent( int i, vtkGenericPointIterator* p );
  
  // Description:
  // Recursive duplication of `other' in `this'.
  // \pre other_exists: other!=0
  // \pre not_self: other!=this
  virtual void DeepCopy( vtkGenericAttribute* other );
  
  // Description:
  // Update `this' using fields of `other'.
  // \pre other_exists: other!=0
  // \pre not_self: other!=this
  virtual void ShallowCopy( vtkGenericAttribute* other );

  // Description:
  // Access the underlying array.
  vtkDoubleArray* GetArray() { return this->Array; }
  virtual void SetArray( vtkDoubleArray* arr );

  // Description:
  // Return the basis used to define the attributes.
  // This enumerates where in the DeRham exact sequence
  // the basis comes from. Cell-constant quantities are
  // first, followed by face attributes (defined using
  // linear functionals dual to the \f$\partial_2\f$ boundary operator),
  // then edge attributes (defined using linear functionals
  // dual to the \f$\partial_1\f$ boundary operator), and finally
  // vertex attributes (defined using linear functionals
  // dual to the \f$\partial_0\f$ boundary operator).
  vtkGetMacro(DeRhamBasis,int);

protected:
  vtkEdgeFaceAttribute();
  ~vtkEdgeFaceAttribute();

  vtkGetObjectMacro(DataSet,vtkEdgeFaceDataSet);
  virtual void SetDataSet( vtkEdgeFaceDataSet* ds );

  // Description:
  // Set GetDeRhamBasis.
  vtkSetMacro(DeRhamBasis,int);

  vtkDoubleArray* Array;
  vtkEdgeFaceDataSet* DataSet;
  int Type;
  int Centering;
  int DeRhamBasis;

  //BTX
  friend class vtkEdgeFaceDataSet;
  friend class vtkEdgeFaceReader;
  //ETX

private:
  vtkEdgeFaceAttribute( const vtkEdgeFaceAttribute& ); // Not implemented.
  void operator = ( const vtkEdgeFaceAttribute& ); // Not implemented.
};

#endif
