#include "vtkEdgeFaceCell.h"

#include "vtkEdgeFaceAttribute.h"
#include "vtkEdgeFaceDataSet.h"

#include <assert.h>

#include "vtkPoints.h"
#include "vtkPointLocator.h"
#include "vtkTetra.h"
#include "vtkTriangle.h"
#include "vtkLine.h"
#include "vtkVertex.h"
#include "vtkPointData.h"
#include "vtkCellData.h"
#include "vtkDoubleArray.h"
#include "vtkCellArray.h"
#include "vtkContourValues.h"
#include "vtkImplicitFunction.h"
#include "vtkGenericAttributeCollection.h"
#include "vtkGenericAttribute.h"
#include "vtkGenericCellTessellator.h"
#include "vtkUnsignedCharArray.h"
#include "vtkQuad.h"
#include "vtkHexahedron.h"
#include "vtkWedge.h"
#include "vtkPyramid.h"
#include "vtkObjectFactory.h"

vtkStandardNewMacro(vtkEdgeFaceCell);
vtkCxxRevisionMacro(vtkEdgeFaceCell, "$Revision: 1.25 $");
vtkCxxSetObjectMacro(vtkEdgeFaceCell,DataSet,vtkEdgeFaceDataSet);

double vtkEdgeFaceCell::ParametricCorners[] =
{
  0., 0., 0.,
  1., 0., 0.,
  1., 1., 0.,
  0., 1., 0.,
  0., 0., 1.,
  1., 0., 1.,
  1., 1., 1.,
  0., 1., 1.
};

#if 0
// This version has faces oriented consistently with the implicit r-s-t axes of the hex.
int vtkEdgeFaceCell::FaceArrays[6][4] =
{
  { 0, 3, 7, 4 }, // -X
  { 1, 2, 6, 5 }, // +X
  { 0, 1, 5, 4 }, // -Y
  { 3, 2, 6, 7 }, // +Y
  { 0, 1, 2, 3 }, // -Z
  { 4, 5, 6, 7 }  // +Z
};

int vtkEdgeFaceCell::EdgeArrays[12][2] =
{
  { 0, 1 }, // X edges
  { 3, 2 },
  { 4, 5 },
  { 7, 6 },
  { 0, 3 }, // Y edges
  { 1, 2 },
  { 4, 7 },
  { 5, 6 },
  { 0, 4 }, // Z edges
  { 1, 5 },
  { 3, 7 },
  { 2, 6 }
};
#else
// This version has faces with inwards-pointing normals and "traditional" edges
int vtkEdgeFaceCell::FaceArrays[6][4] =
{
  { 0, 3, 7, 4 },
  { 1, 5, 6, 2 },
  { 0, 1, 5, 4 },
  { 3, 7, 6, 2 },
  { 0, 1, 2, 3 },
  { 4, 7, 6, 5 }
};

int vtkEdgeFaceCell::EdgeArrays[12][2] =
{
  { 0, 1 },
  { 1, 2 },
  { 2, 3 },
  { 3, 0 },
  { 0, 4 },
  { 1, 5 },
  { 3, 7 },
  { 2, 6 },
  { 4, 5 },
  { 5, 6 },
  { 6, 7 },
  { 7, 4 }
};
#endif // 0

vtkEdgeFaceCell::vtkEdgeFaceCell()
{
  this->DataSet = 0;
  this->Dimension = 3;
  this->Id = -1;
}

vtkEdgeFaceCell::~vtkEdgeFaceCell()
{
  this->SetDataSet( 0 );
}

void vtkEdgeFaceCell::PrintSelf( ostream& os, vtkIndent indent )
{
  this->Superclass::PrintSelf( os, indent );
  os << indent << "DataSet: " << this->DataSet << "\n";
  os << indent << "Dimension: " << this->Dimension << "\n";
  os << indent << "Id: " << this->Id << "\n";
}

vtkIdType vtkEdgeFaceCell::GetId()
{
  return this->Id;
}

int vtkEdgeFaceCell::IsInDataSet()
{
  // Since the mesh only contains hexahedra, we can just check the dimension of the cell.
  // If it's not 3, then we're representing a cell boundary.
  return this->Dimension == 3;
}

int vtkEdgeFaceCell::GetType()
{
  switch ( this->Dimension )
    {
  case 3:
    return VTK_HIGHER_ORDER_HEXAHEDRON;
  case 2:
    return VTK_HIGHER_ORDER_QUAD;
  case 1:
    return VTK_HIGHER_ORDER_EDGE;
  default:
    return VTK_EMPTY_CELL;
    }
}

int vtkEdgeFaceCell::GetDimension()
{
  return this->Dimension;
}

int vtkEdgeFaceCell::GetGeometryOrder()
{
  return 1;
}

int vtkEdgeFaceCell::GetAttributeOrder( vtkGenericAttribute* a )
{
  (void) a; // All attributes are linear, so we can ignore a
  return 1;
}

int vtkEdgeFaceCell::IsPrimary()
{
  // Our mesh only holds hexahedra.
  // If you have a mesh that contains strips or fans of elements
  // in a single cell, this should return 0 (false) for those cells.
  return 1;
}

int vtkEdgeFaceCell::GetNumberOfPoints()
{
  switch ( this->Dimension )
    {
  case 3:
    return 8;
  case 2:
    return 4;
  case 1:
    return 2;
  default:
    return 0;
    }
}

int vtkEdgeFaceCell::GetNumberOfBoundaries( int dim )
{
  switch ( this->Dimension )
    {
  case 3:
    switch ( dim )
      {
    case 2:
      return 6;
    case 1:
      return 12;
    case 0:
      return 8;
    default:
      return 0;
      }
  case 2:
    switch ( dim )
      {
    case 1:
    case 0:
      return 4;
    default:
      return 0;
      }
  case 1:
    return dim ? 0 : 2;
  default:
    return 0;
    }
}

int vtkEdgeFaceCell::GetNumberOfDOFNodes()
{
  switch ( this->Dimension )
    {
  case 3:
    return 12 + 6;
  case 2:
    return 4;
  case 1:
    return 1;
  default:
    return 0;
    }
}

void vtkEdgeFaceCell::GetPointIterator( vtkGenericPointIterator* it )
{
  (void) it;
  // FIXME
}

vtkGenericCellIterator* vtkEdgeFaceCell::NewCellIterator()
{
  // FIXME
  return 0;
}

void vtkEdgeFaceCell::GetBoundaryIterator( vtkGenericCellIterator* boundaries, int dim )
{
  (void) boundaries;
  (void) dim;
}

int vtkEdgeFaceCell::CountNeighbors( vtkGenericAdaptorCell* boundary )
{
  (void) boundary;
  // FIXME
  return -1;
}

void vtkEdgeFaceCell::CountEdgeNeighbors( int* sharing )
{
  (void) sharing;
  // FIXME
}

void vtkEdgeFaceCell::GetNeighbors( vtkGenericAdaptorCell* boundary, vtkGenericCellIterator* neighbors )
{
  (void) boundary;
  (void) neighbors;
  // FIXME
}

int vtkEdgeFaceCell::FindClosestBoundary( int subId, double pcoords[3], vtkGenericCellIterator* &boundary )
{
  (void) subId;
  (void) pcoords;
  (void) boundary;
  // FIXME
  return -1;
}

int vtkEdgeFaceCell::EvaluatePosition( double x[3], double* closestPoint, int &subId, double pcoords[3], double &dist2 )
{
  (void) x;
  (void) closestPoint;
  (void) subId;
  (void) pcoords;
  (void) dist2;
  // FIXME
  return -1;
}

void vtkEdgeFaceCell::EvaluateLocation( int subId, double pcoords[3], double x[3] )
{
  (void) subId; // Ignore the subid since all cells are hexahedra.
  int i;
  int j;

  double cellpts[3];
  vtkCellArray* cellArr = this->DataSet->GetConnectivity();
  vtkIdType* conn;
  vtkIdType npts;
  cellArr->GetCell( 27 * this->Id, npts, conn );
  for ( i = 0; i < 8; ++ i )
    {
    }
  x[0] = x[1] = x[2] = 0.;

  vtkPoints* pts = this->DataSet->GetPoints();
  for ( i = 0; i < 8; ++ i )
    {
    double wt =
      ( ( ( i & 0x3 ) == 0x3 || ( i & 0x3 ) == 0x0 ) ? ( 1. - pcoords[0] ) : pcoords[0] ) *
      (   ( i & 0x2 ) ? pcoords[1] : ( 1. - pcoords[1] ) ) *
      (   ( i & 0x4 ) ? pcoords[2] : ( 1. - pcoords[2] ) );
    pts->GetPoint( conn[i], cellpts );
    for ( j = 0; j < 3; ++ j )
      {
      x[j] += cellpts[j] * wt;
      }
    }
}

void vtkEdgeFaceCell::InterpolateTuple( vtkGenericAttribute* a, double pcoords[3], double* val )
{
  vtkEdgeFaceAttribute* att = vtkEdgeFaceAttribute::SafeDownCast( a );
  if ( a )
    {
    vtkPoints* pts = this->DataSet->GetPoints();
    double cellatts[12]; // at most 12 values per hex (may be 1, 6, 8, or 12)
    double cellpts[8][3];
    double vec[3][3];
    int i;
    vtkEdgeFaceAttribute::DeRhamBasisValues basis = (vtkEdgeFaceAttribute::DeRhamBasisValues) att->GetDeRhamBasis();
    vtkDoubleArray* attArr = vtkDoubleArray::SafeDownCast( att->GetArray() );
    vtkCellArray* cellArr = this->DataSet->GetConnectivity();
    vtkIdType* conn;
    vtkIdType npts;
    cellArr->GetCell( 27 * this->Id, npts, conn );
    for ( i = 0; i < 8; ++ i )
      {
      pts->GetPoint( conn[i], cellpts[i] );
      }
    switch ( basis )
      {
    case vtkEdgeFaceAttribute::CELL:
        {
        this->ComputeCellShapeVectors( &vec[0][0], pcoords, &cellpts[0][0] );
        double shapeCell;
        cellatts[0] = attArr->GetValue( conn[8 + 12 + 6] );
        this->GetCellBasis( &shapeCell, &vec[0][0], pcoords );
        *val = shapeCell * cellatts[0];
        }
      break;
    case vtkEdgeFaceAttribute::FACE:
        {
        this->ComputeCellShapeVectors( &vec[0][0], pcoords, &cellpts[0][0] );
        double shapeFace[6 * 3];
        for ( i = 0; i < 6; ++ i )
          {
          cellatts[i] = attArr->GetValue( conn[i + 8 + 12] );
          }
        double* sfp = shapeFace;
        this->GetFaceBasis( shapeFace, &vec[0][0], pcoords );
        for ( i = 0; i < 3; ++ i )
          {
          val[i] = 0.;
          }
        for ( i = 0; i < 6; ++ i, sfp += 3 )
          {
          for ( int j = 0; j < 3; ++ j )
            {
            val[j] += cellatts[i] * sfp[j];
            }
          }
        }
      break;
    case vtkEdgeFaceAttribute::EDGE:
        {
        this->ComputeCellShapeVectors( &vec[0][0], pcoords, &cellpts[0][0] );
        double shapeEdge[12 * 3];
        for ( i = 0; i < 12; ++ i )
          {
          cellatts[i] = attArr->GetValue( conn[i + 8] );
          }
        double* sep = shapeEdge;
        this->GetEdgeBasis( shapeEdge, &vec[0][0], pcoords );
        for ( i = 0; i < 3; ++ i )
          {
          val[i] = 0.;
          }
        for ( i = 0; i < 12; ++ i, sep += 3 )
          {
          for ( int j = 0; j < 3; ++ j )
            {
            val[j] += cellatts[i] * sep[j];
            }
          }
        }
      break;
    case vtkEdgeFaceAttribute::NODE:
        {
        this->ComputeCellShapeVectors( &vec[0][0], pcoords, &cellpts[0][0] );
        double shapeNode[8];
        for ( i = 0; i < 8; ++ i )
          {
          cellatts[i] = attArr->GetValue( conn[i] );
          }
        double* snp = shapeNode;
        this->GetNodeBasis( shapeNode, &vec[0][0], pcoords );
        *val = cellatts[0] * snp[0];
        for ( i = 1; i < 8; ++ i )
          {
          *val += cellatts[i] * snp[i];
          }
        }
      break;
      }
    }
}

void vtkEdgeFaceCell::InterpolateTuple( vtkGenericAttributeCollection* c, double pcoords[3], double* val )
{
  //int n = c->GetNumberOfAttributesToInterpolate();
  //int* alist = c->GetAttributesToInterpolate();
  int n = c->GetNumberOfAttributes();
  int offset = 0;
  for ( int i = 0; i < n; ++ i )
    {
    vtkEdgeFaceAttribute* att = vtkEdgeFaceAttribute::SafeDownCast( c->GetAttribute( i ) );
    if ( ! att )
      {
      continue;
      }
    this->InterpolateTuple( att, pcoords, val + offset );
    offset += att->GetNumberOfComponents();
    }
}

double* vtkEdgeFaceCell::GetAttributeAtCorners( vtkEdgeFaceAttribute* a )
{
  this->AllocateTuples( 8 * 3 );
  double* pcorners = vtkEdgeFaceCell::GetParametricCorners();
  double* acorners = this->Tuples;
  for ( int i = 0; i < 8; ++ i, pcorners += 3, acorners += 3 )
    {
    this->InterpolateTuple( a, pcorners, acorners );
    }
  return this->Tuples;
}

void vtkEdgeFaceCell::Contour(
  vtkContourValues* contourValues, vtkImplicitFunction* f,
  vtkGenericAttributeCollection* attributes, vtkGenericCellTessellator* tess,
  vtkPointLocator* locator, vtkCellArray* verts, vtkCellArray* lines,
  vtkCellArray* polys, vtkPointData* outPd, vtkCellData* outCd,
  vtkPointData* internalPd, vtkPointData* secondaryPd,
  vtkCellData* secondaryCd )
{
  this->Superclass::Contour( contourValues, f, attributes, tess, locator,
    verts, lines, polys, outPd, outCd, internalPd, secondaryPd, secondaryCd );
}

void vtkEdgeFaceCell::Clip(
  double value, vtkImplicitFunction* f,
  vtkGenericAttributeCollection* attributes, vtkGenericCellTessellator* tess,
  int insideOut, vtkPointLocator* locator, vtkCellArray* connectivity,
  vtkPointData* outPd, vtkCellData* outCd, vtkPointData* internalPd,
  vtkPointData* secondaryPd, vtkCellData* secondaryCd )
{
  this->Superclass::Clip( value, f, attributes, tess, insideOut, locator,
    connectivity, outPd, outCd, internalPd, secondaryPd, secondaryCd );
}

int vtkEdgeFaceCell::IntersectWithLine( double p1[3], double p2[3], double tol, double &t, double x[3], double pcoords[3], int &subId )
{
  (void) p1;
  (void) p2;
  (void) tol;
  (void) t;
  (void) x;
  (void) pcoords;
  (void) subId;
  return -1; // FIXME
}

void vtkEdgeFaceCell::Derivatives( int subId, double pcoords[3], vtkGenericAttribute* attribute, double* derivs )
{
  (void) subId;
  (void) pcoords;
  (void) attribute;
  (void) derivs;
  // FIXME
}


void vtkEdgeFaceCell::GetBounds( double bounds[6] )
{
  this->ComputeBounds( bounds );
  // Maybe one day we'll store the results in the cell so multiple calls
  // to GetBounds() won't recompute bounds.
}

double* vtkEdgeFaceCell::GetBounds()
{
  static double bds[6];
  this->GetBounds( bds );
  return bds;
}

int vtkEdgeFaceCell::GetParametricCenter( double pcoords[3] )
{
  (void) pcoords;
  return 0; // FIXME
}

double vtkEdgeFaceCell::GetParametricDistance( double pcoords[3] )
{
  (void) pcoords;
  return -1.; // FIXME
}

double* vtkEdgeFaceCell::GetParametricCoords()
{
  return vtkEdgeFaceCell::ParametricCorners;
}


void vtkEdgeFaceCell::Tessellate(
  vtkGenericAttributeCollection* attributes, vtkGenericCellTessellator* tess,
  vtkPoints* points, vtkPointLocator* locator, vtkCellArray* cellArray,
  vtkPointData* internalPd, vtkPointData* pd, vtkCellData* cd,
  vtkUnsignedCharArray* types )
{
  this->Superclass::Tessellate( attributes, tess, points, locator, cellArray,
    internalPd, pd, cd, types );
}

int vtkEdgeFaceCell::IsFaceOnBoundary( vtkIdType faceId )
{
  if (
    faceId < 0 || faceId >= 6 ||
    this->Id < 0 || this->Id >= this->DataSet->GetNumberOfCells() )
    {
    return 0;
    }

  return this->DataSet->GetNeighborFaces()->GetComponent( this->Id, faceId ) == 255;
}

int vtkEdgeFaceCell::IsOnBoundary()
{
  return -1; // FIXME
}

void vtkEdgeFaceCell::GetPointIds( vtkIdType* id )
{
  (void) id;
}


void vtkEdgeFaceCell::TriangulateFace(
  vtkGenericAttributeCollection* attributes, vtkGenericCellTessellator* tess,
  int index, vtkPoints* points, vtkPointLocator* locator,
  vtkCellArray* cellArray, vtkPointData* internalPd,
  vtkPointData* pd, vtkCellData* cd )
{
  this->Superclass::TriangulateFace( attributes, tess, index, points, locator,
    cellArray, internalPd, pd, cd );
}

int* vtkEdgeFaceCell::GetFaceArray( int faceId )
{
  if ( faceId < 0 || faceId >= 6 )
    {
    return 0;
    }
  return vtkEdgeFaceCell::FaceArrays[faceId];
}

int vtkEdgeFaceCell::GetNumberOfVerticesOnFace( int faceId )
{
  (void) faceId; // Always 4
  return 4;
}

int* vtkEdgeFaceCell::GetEdgeArray( int edgeId )
{
  if ( edgeId < 0 || edgeId >= 12 )
    {
    return 0;
    }
  return vtkEdgeFaceCell::EdgeArrays[edgeId];
}

double* vtkEdgeFaceCell::GetParametricCorners()
{
  return vtkEdgeFaceCell::ParametricCorners;
}

void vtkEdgeFaceCell::Reset()
{
  this->Superclass::Reset();
}

void vtkEdgeFaceCell::ComputeCellShapeVectors( double* vec, double pcoords[3], double* cellpts )
{
  int i;
  double* v1 = vec;
  double* v2 = vec + 3;
  double* v3 = vec + 6;
  double lmn = 1. - pcoords[0];
  double lmp = 1. + pcoords[0];
  double mun = 1. - pcoords[1];
  double mup = 1. + pcoords[1];
  double nun = 1. - pcoords[2];
  double nup = 1. + pcoords[2];
  for ( i = 0; i < 3; ++ i )
    {
    v1[i] = 0.125 * (
      nun * ( mun * ( cellpts[3 + i] - cellpts[0 + i] ) +
              mup * ( cellpts[6 + i] - cellpts[9 + i] ) ) +
      nup * ( mun * ( cellpts[15 + i] - cellpts[12 + i] ) +
              mup * ( cellpts[18 + i] - cellpts[21 + i] ) ) );
    v2[i] = 0.125 * (
      nun * ( lmn * ( cellpts[9 + i] - cellpts[0 + i] ) +
              lmp * ( cellpts[6 + i] - cellpts[3 + i] ) ) +
      nup * ( lmn * ( cellpts[21 + i] - cellpts[12 + i] ) +
              lmp * ( cellpts[18 + i] - cellpts[15 + i] ) ) );
    v3[i] = 0.125 * (
      mun * ( lmn * ( cellpts[12 + i] - cellpts[0 + i] ) +
              lmp * ( cellpts[15 + i] - cellpts[3 + i] ) ) +
      mup * ( lmn * ( cellpts[18 + i] - cellpts[6 + i] ) +
              lmp * ( cellpts[21 + i] - cellpts[9 + i] ) ) );
    }
}

double vtkEdgeFaceCell::DetCrossJF( double* vec, double v1xv2[3], double v1xv3[3], double v2xv3[3] )
{
  double* v1 = vec;
  double* v2 = vec + 3;
  double* v3 = vec + 6;

  v1xv2[0] = v1[1] * v2[2] - v1[2] * v2[1];
  v1xv2[1] = v1[2] * v2[0] - v1[0] * v2[2];
  v1xv2[2] = v1[0] * v2[1] - v1[1] * v2[0];

  v1xv3[0] = v1[1] * v3[2] - v1[2] * v3[1];
  v1xv3[1] = v1[2] * v3[0] - v1[0] * v3[2];
  v1xv3[2] = v1[0] * v3[1] - v1[1] * v3[0];

  v2xv3[0] = v2[1] * v3[2] - v2[2] * v3[1];
  v2xv3[1] = v2[2] * v3[0] - v2[0] * v3[2];
  v2xv3[2] = v2[0] * v3[1] - v2[1] * v3[0];

  // Compute the determinant 3 ways and average the results.
  double detJFa = v1[0] * v2xv3[0] + v1[1] * v2xv3[1] + v1[2] * v2xv3[2];
  double detJFb = - ( v2[0] * v1xv3[0] + v2[1] * v1xv3[1] + v2[2] * v1xv3[2] );
  double detJFc = v3[0] * v1xv2[0] + v3[1] * v1xv2[1] + v3[2] * v1xv2[2];

  return ( detJFa + detJFb + detJFc ) / 3.;
}

void vtkEdgeFaceCell::GetNodeBasis( double* shapeNode, double* vec, double pcoords[3] )
{
  (void) vec; // Not used by the linear nodal basis.
  for ( int m = 0; m < 3; ++ m )
    {
    shapeNode[ 0 + m] = 0.125 * ( 1. - pcoords[0] ) * ( 1. - pcoords[1] ) * ( 1. - pcoords[2] );
    shapeNode[ 3 + m] = 0.125 * ( 1. + pcoords[0] ) * ( 1. - pcoords[1] ) * ( 1. - pcoords[2] );
    shapeNode[ 6 + m] = 0.125 * ( 1. + pcoords[0] ) * ( 1. + pcoords[1] ) * ( 1. - pcoords[2] );
    shapeNode[ 9 + m] = 0.125 * ( 1. - pcoords[0] ) * ( 1. + pcoords[1] ) * ( 1. - pcoords[2] );

    shapeNode[12 + m] = 0.125 * ( 1. - pcoords[0] ) * ( 1. - pcoords[1] ) * ( 1. + pcoords[2] );
    shapeNode[15 + m] = 0.125 * ( 1. + pcoords[0] ) * ( 1. - pcoords[1] ) * ( 1. + pcoords[2] );
    shapeNode[18 + m] = 0.125 * ( 1. + pcoords[0] ) * ( 1. + pcoords[1] ) * ( 1. + pcoords[2] );
    shapeNode[21 + m] = 0.125 * ( 1. - pcoords[0] ) * ( 1. + pcoords[1] ) * ( 1. + pcoords[2] );
    }
}

void vtkEdgeFaceCell::GetEdgeBasis( double* shapeEdge, double* vec, double pcoords[3] )
{
  double vlmmu[3];
  double vlmnu[3];
  double vmunu[3];
  double detJF = this->DetCrossJF( vec, vlmmu, vlmnu, vmunu );

  for ( int m = 0; m < 3; ++ m )
    {
    shapeEdge[ 0 + m] = 0.125 * ( 1. - pcoords[1] ) * ( 1. - pcoords[2] ) * vmunu[m] / detJF;
    shapeEdge[ 3 + m] = 0.125 * ( 1. + pcoords[0] ) * ( 1. - pcoords[2] ) * vlmnu[m] / detJF;
    shapeEdge[ 6 + m] = 0.125 * ( 1. + pcoords[1] ) * ( 1. - pcoords[2] ) * vmunu[m] / detJF;
    shapeEdge[ 9 + m] = 0.125 * ( 1. - pcoords[0] ) * ( 1. - pcoords[2] ) * vlmnu[m] / detJF;

    shapeEdge[12 + m] = 0.125 * ( 1. - pcoords[0] ) * ( 1. - pcoords[1] ) * vlmmu[m] / detJF;
    shapeEdge[15 + m] = 0.125 * ( 1. + pcoords[0] ) * ( 1. - pcoords[1] ) * vlmmu[m] / detJF;
    shapeEdge[18 + m] = 0.125 * ( 1. - pcoords[0] ) * ( 1. + pcoords[1] ) * vlmmu[m] / detJF;
    shapeEdge[21 + m] = 0.125 * ( 1. + pcoords[0] ) * ( 1. + pcoords[1] ) * vlmmu[m] / detJF;

    shapeEdge[24 + m] = 0.125 * ( 1. - pcoords[1] ) * ( 1. + pcoords[2] ) * vmunu[m] / detJF;
    shapeEdge[27 + m] = 0.125 * ( 1. + pcoords[0] ) * ( 1. + pcoords[2] ) * vlmnu[m] / detJF;
    shapeEdge[30 + m] = 0.125 * ( 1. + pcoords[1] ) * ( 1. + pcoords[2] ) * vmunu[m] / detJF;
    shapeEdge[33 + m] = 0.125 * ( 1. - pcoords[0] ) * ( 1. + pcoords[2] ) * vlmnu[m] / detJF;
    }
}

void vtkEdgeFaceCell::GetFaceBasis( double* shapeFace, double* vec, double pcoords[3] )
{
  double vlmmu[3];
  double vlmnu[3];
  double vmunu[3];
  double detJF = this->DetCrossJF( vec, vlmmu, vlmnu, vmunu );

  for ( int m = 0; m < 3; ++ m )
    {
    double fac;
    fac = 0.125 * vec[m] / detJF;
    shapeFace[ 0 + m] = fac * ( 1. - pcoords[0] );
    shapeFace[ 3 + m] = fac * ( 1. + pcoords[0] );

    fac = 0.125 * vec[m + 3] / detJF;
    shapeFace[ 6 + m] = fac * ( 1. + pcoords[1] );
    shapeFace[ 9 + m] = fac * ( 1. - pcoords[1] );

    fac = 0.125 * vec[m + 6] / detJF;
    shapeFace[12 + m] = fac * ( 1. - pcoords[2] );
    shapeFace[15 + m] = fac * ( 1. + pcoords[2] );
    }
}

void vtkEdgeFaceCell::GetCellBasis( double* shapeCell, double* vec, double pcoords[3] )
{
  (void) pcoords; // The cell basis is constant over the entire cell so pcoords is irrelevant.
  double vlmmu[3];
  double vlmnu[3];
  double vmunu[3];
  double detJF = this->DetCrossJF( vec, vlmmu, vlmnu, vmunu );

  *shapeCell = 0.125 * ( vec[0] * vmunu[0] + vec[1] * vmunu[1] + vec[2] * vmunu[2] ) / detJF;
}

void vtkEdgeFaceCell::ComputeBounds( double bds[6] )
{
  double corner[3];
  vtkIdType* conn;
  vtkIdType npts;
  int i;
  vtkCellArray* cellArr = this->DataSet->GetConnectivity();
  cellArr->GetCell( 27 * this->Id, npts, conn );
  this->DataSet->GetPoint( conn[0], corner );
  for ( i = 0; i < 3; ++ i )
    bds[2 * i] = bds[2 * i + 1] = corner[i];
  for ( i = 1; i < 8; ++ i )
    {
    this->DataSet->GetPoint( conn[i], corner );
    }
}
