#include "vtkEdgeFaceDataSet.h"

#include "vtkEdgeFaceCell.h"
#include "vtkEdgeFaceCellIterator.h"
#include "vtkEdgeFaceLocator.h"
#include "vtkEdgeFacePointIterator.h"

#include "vtkCellTypes.h"
#include "vtkCellArray.h"
#include "vtkGenericAdaptorCell.h"
#include "vtkGenericAttributeCollection.h"
#include "vtkGenericCellTessellator.h"
#include "vtkSimpleCellTessellator.h"
#include "vtkIdTypeArray.h"
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkObjectFactory.h"
#include "vtkPoints.h"
#include "vtkUnsignedCharArray.h"

#include <vtkstd/set>
#include <vtkstd/map>

#include <assert.h>

vtkStandardNewMacro(vtkEdgeFaceDataSet);
vtkCxxRevisionMacro(vtkEdgeFaceDataSet, "$Revision: 1.8 $");
vtkCxxSetObjectMacro(vtkEdgeFaceDataSet,Connectivity,vtkCellArray);
vtkCxxSetObjectMacro(vtkEdgeFaceDataSet,Points,vtkPoints);
vtkCxxSetObjectMacro(vtkEdgeFaceDataSet,Locator,vtkEdgeFaceLocator);
vtkCxxSetObjectMacro(vtkEdgeFaceDataSet,NeighborIds,vtkIdTypeArray);
vtkCxxSetObjectMacro(vtkEdgeFaceDataSet,NeighborFaces,vtkUnsignedCharArray);
vtkCxxSetObjectMacro(vtkEdgeFaceDataSet,CellLookup,vtkIdTypeArray);

vtkEdgeFaceDataSet::vtkEdgeFaceDataSet()
{
  this->Points = vtkPoints::New();
  this->Connectivity = vtkCellArray::New();
  this->Locator = 0;
  this->NeighborIds = vtkIdTypeArray::New();
  this->NeighborFaces = vtkUnsignedCharArray::New();
  this->CellLookup = vtkIdTypeArray::New();
  this->NeighborIds->SetNumberOfComponents( 6 );
  this->NeighborFaces->SetNumberOfComponents( 6 );
  this->CellLookup->SetNumberOfComponents( 1 );
  vtkGenericCellTessellator* tess = vtkSimpleCellTessellator::New();
  this->SetTessellator( tess );
  tess->Delete();
}

vtkEdgeFaceDataSet::~vtkEdgeFaceDataSet()
{
  this->SetPoints( 0 );
  this->SetConnectivity( 0 );
  this->SetLocator( 0 );
  this->SetNeighborIds( 0 );
  this->SetNeighborFaces( 0 );
  this->SetCellLookup( 0 );
}

void vtkEdgeFaceDataSet::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);
  os << indent << "Points: " << this->Points << "\n";
  os << indent << "Connectivity: " << this->Connectivity << "\n";
  os << indent << "Locator: " << this->Locator << "\n";
  os << indent << "NeighborIds: " << this->NeighborIds << "\n";
  os << indent << "NeighborFaces: " << this->NeighborFaces << "\n";
  os << indent << "CellLookup: " << this->CellLookup << "\n";
}

vtkIdType vtkEdgeFaceDataSet::GetNumberOfPoints()
{
  return this->Points->GetNumberOfPoints();
}

vtkIdType vtkEdgeFaceDataSet::GetNumberOfCells( int dim )
{
  if ( dim == 3 || dim == -1 )
    {
    return this->Connectivity->GetNumberOfCells();
    }
  return 0;
}

int vtkEdgeFaceDataSet::GetCellDimension()
{
  // We only store hexahedra.
  return 3;
}

vtkGenericCellIterator* vtkEdgeFaceDataSet::NewCellIterator( int dim )
{
  vtkEdgeFaceCellIterator* cellIt = vtkEdgeFaceCellIterator::New();
  cellIt->SetDataSet( this );
  cellIt->SetDimension( dim );
  return cellIt;
}

vtkGenericCellIterator* vtkEdgeFaceDataSet::NewBoundaryIterator( int dim, int exteriorOnly )
{
  vtkEdgeFaceCellIterator* cellIt = vtkEdgeFaceCellIterator::New();
  cellIt->SetDataSet( this );
  cellIt->SetDimension( dim );
  cellIt->SetMode( exteriorOnly ? vtkEdgeFaceCellIterator::BOUNDARY : vtkEdgeFaceCellIterator::COFACETS );
  return cellIt;
}

vtkGenericPointIterator* vtkEdgeFaceDataSet::NewPointIterator()
{
  vtkEdgeFacePointIterator* ptIt = vtkEdgeFacePointIterator::New();
  ptIt->SetMode( vtkEdgeFacePointIterator::MESH );
  ptIt->SetDataSet( this );
  return ptIt;
}

int vtkEdgeFaceDataSet::FindCell(
  double x[3], vtkGenericCellIterator*& cell, double tol2, int& subId, double pcoords[3] )
{
  vtkEdgeFaceCellIterator* efcit = vtkEdgeFaceCellIterator::SafeDownCast( cell );
  if ( ! efcit )
    {
    return 0;
    }

  vtkEdgeFacePointIterator* ptIt = vtkEdgeFacePointIterator::SafeDownCast( this->NewPointIterator() );
  this->FindPoint( x, ptIt );
  //subId = 0; // Hexahedra are not composite cells, so this is always the case.
  vtkEdgeFaceCell* efcell = vtkEdgeFaceCell::SafeDownCast( efcit->GetCell() );
  double closestPoint[3];
  double dist2;
  int isInCell = efcell->EvaluatePosition( x, closestPoint, subId, pcoords, dist2 );
  if ( isInCell < 0 || dist2 > tol2 )
    {
    return 0;
    }
  return 1; // 0 == point is not in cell, 1 == point is inside cell
}

void vtkEdgeFaceDataSet::FindPoint( double x[3], vtkGenericPointIterator* p )
{
  if ( ! p )
    {
    return;
    }

  vtkEdgeFacePointIterator* efp = vtkEdgeFacePointIterator::SafeDownCast( p );
  assert( efp );

  if ( ! this->Locator )
    {
    this->Locator = vtkEdgeFaceLocator::New();
    this->Locator->SetDataSet( this );
    }

  if ( this->Points->GetMTime() > this->Locator->GetMTime() )
    {
    this->Locator->SetDataSet( this );
    }

  vtkIdType ptId = this->Locator->FindClosestPoint( x );
  efp->SetId( ptId );
}

void vtkEdgeFaceDataSet::ComputeBounds()
{
  this->Points->GetBounds( this->Bounds );
}

unsigned long vtkEdgeFaceDataSet::GetActualMemorySize()
{
  unsigned long result = this->Superclass::GetActualMemorySize();
  result += this->Points->GetActualMemorySize();
  result += this->Connectivity->GetActualMemorySize();
  //result += this->Locator->GetActualMemorySize(); // FIXME: Need size of locator.
  result += this->NeighborIds->GetActualMemorySize();
  result += this->NeighborFaces->GetActualMemorySize();
  result += this->CellLookup->GetActualMemorySize();
  return result;
}

int vtkEdgeFaceDataSet::GetDataObjectType()
{
  // You may want to change this, but leaving as-is should not normally
  // be a problem. It will only matter when a descendent of vtkAlgorithm
  // must be able to differentiate between your subclass and the parent.
  return VTK_GENERIC_DATA_SET;
}

vtkIdType vtkEdgeFaceDataSet::GetEstimatedSize()
{
  return 6 * this->GetNumberOfCells(); // 6 tetrahedra per hex
}

vtkEdgeFaceDataSet* vtkEdgeFaceDataSet::GetData(vtkInformation* info)
{
  return info ? vtkEdgeFaceDataSet::SafeDownCast( info->Get( DATA_OBJECT() ) ) : 0;
}

vtkEdgeFaceDataSet* vtkEdgeFaceDataSet::GetData( vtkInformationVector* v, int i )
{
  return vtkEdgeFaceDataSet::GetData( v->GetInformationObject( i ) );
}

void vtkEdgeFaceDataSet::GetPoint( vtkIdType pt, double* x )
{
  this->Points->GetPoint( pt, x );
}

double* vtkEdgeFaceDataSet::GetPoint( vtkIdType pt )
{
  static double coords[3];
  this->GetPoint( pt, coords );
  return coords;
}

struct vtkEdgeFaceQuadRec
{
  vtkIdType Connectivity[4];
  vtkEdgeFaceQuadRec( vtkIdType* hexConn, int* faceIndices )
    {
    vtkstd::set<vtkIdType> x;
    int i;
    for ( i = 0; i < 4; ++ i )
      x.insert( hexConn[faceIndices[i]] );
    i = -1;
    for ( vtkstd::set<vtkIdType>::iterator it = x.begin(); it != x.end(); ++ it )
      {
      this->Connectivity[++ i] = *it;
      }
    }
  bool operator < ( const vtkEdgeFaceQuadRec& other ) const
    {
    for ( int i = 0; i < 4; ++ i )
      {
      if ( this->Connectivity[i] < other.Connectivity[i] )
        return true;
      else if ( this->Connectivity[i] > other.Connectivity[i] )
        return false;
      }
    return false;
    }
  bool operator == ( const vtkEdgeFaceQuadRec& other ) const
    {
    for ( int i = 0; i < 4; ++ i )
      {
      if ( this->Connectivity[i] != other.Connectivity[i] )
        return false;
      }
    return true;
    }
};

int vtkEdgeFaceDataSetFindSharedFace(
  vtkEdgeFaceQuadRec& frec, vtkIdType* oconn )
{
  int oidx;
  for ( oidx = 0; oidx < 6; ++ oidx )
    {
    vtkEdgeFaceQuadRec ofrec( oconn, vtkEdgeFaceCell::FaceArrays[oidx] );
    if ( frec == ofrec )
      {
      return oidx;
      }
    }
  return -1;
}

void vtkEdgeFaceDataSet::ComputeNeighbors()
{
  this->NeighborIds->SetNumberOfTuples( this->Connectivity->GetNumberOfCells() );
  this->NeighborFaces->SetNumberOfTuples( this->Connectivity->GetNumberOfCells() );
  this->CellLookup->SetNumberOfTuples( this->Points->GetNumberOfPoints() );
  this->CellLookup->FillComponent( 0, -1. );
  vtkIdType* conn;
  vtkIdType npts;
  vtkIdType i;
  vtkIdType* lkup = this->CellLookup->GetPointer( 0 );
  vtkIdType* nbor = this->NeighborIds->GetPointer( 0 );
  unsigned char* face = this->NeighborFaces->GetPointer( 0 );
  vtkstd::map<vtkEdgeFaceQuadRec,vtkIdType> faceMap;
  int j;
  for ( i = 0; i < this->Connectivity->GetNumberOfCells(); ++ i )
    {
    this->Connectivity->GetCell( 27 * i, npts, conn );
    // First, update point->cell lookup table
    for ( j = 0; j < 8; ++ j )
      {
      if ( lkup[conn[j]] < 0 )
        {
        lkup[conn[j]] = i;
        }
      }
    // Now update nbor and face arrays
    for ( j = 0; j < 6; ++ j )
      {
      vtkEdgeFaceQuadRec frec( conn, vtkEdgeFaceCell::FaceArrays[j] );
      vtkstd::map<vtkEdgeFaceQuadRec,vtkIdType>::iterator it;
      it = faceMap.find( frec );
      if ( it == faceMap.end() )
        {
        faceMap[frec] = i;
        nbor[6*i + j] = -1;
        face[6*i + j] = 255;
        }
      else
        {
        vtkIdType other = it->second;
        faceMap.erase( it );
        int oidx; // face index on element other neighboring element i face j
        vtkIdType* oconn;
        this->Connectivity->GetCell( 27 * other, npts, oconn );
        oidx = vtkEdgeFaceDataSetFindSharedFace( frec, oconn );
        if ( oidx >= 0 )
          {
          nbor[6*other + oidx] = i;
          face[6*other + oidx] = j;
          nbor[6*i + j] = other;
          face[6*i + j] = oidx;
          }
        else
          {
          vtkErrorMacro(
            "Cell " << other << " and " << i << " are supposed to share a face with vertices {"
            << frec.Connectivity[0] << "," << frec.Connectivity[1] << ","
            << frec.Connectivity[2] << "," << frec.Connectivity[3] << "} but do not appear to..."
            );
          nbor[6*i + j] = -1;
          face[6*i + j] = 255;
          }
        }
      }
    }
}

