// .NAME vtkEdgeFaceDataSet - defines dataset interface
// .SECTION Description
// In VTK, spatial-temporal data is defined in terms of a dataset. The
// dataset consists of geometry (e.g., points), topology (e.g., cells), and
// attributes (e.g., scalars, vectors, etc.) vtkGenericDataSet is an abstract
// class defining this abstraction.
//
// Since vtkGenericDataSet provides a general interface to manipulate data,
// algorithms that process it tend to be slower than those specialized for a
// particular data type. For this reason, there are concrete, non-abstract
// subclasses that represent and provide access to data more efficiently.
// Note that filters to process this dataset type are currently found in the
// VTK/GenericFiltering/ subdirectory.
//
// Unlike the vtkDataSet class, vtkGenericDataSet provides a more flexible
// interface including support for iterators. vtkGenericDataSet is also
// designed to interface VTK to external simulation packages without the
// penalty of copying memory (see VTK/GenericFiltering/README.html) for
// more information. Thus vtkGenericDataSet plays a central role in the
// adaptor framework.
//
// Please note that this class introduces the concepts of "boundary cells".
// This refers to the boundaries of a cell (e.g., face of a tetrahedron)
// which may in turn be represented as a cell. Boundary cells are derivative
// topological features of cells, and are therefore never explicitly
// represented in the dataset. Often in visualization algorithms, looping
// over boundaries (edges or faces) is employed, while the actual dataset
// cells may not traversed. Thus there are methods to loop over these
// boundary cells.
//
// Finally, as a point of clarification, points are not the same as vertices.
// Vertices refer to points, and points specify a position is space. Vertices
// are a type of 0-D cell. Also, the concept of a DOFNode, which is where
// coefficients for higher-order cells are kept, is a new concept introduced
// by the adaptor framework (see vtkGenericAdaptorCell for more information).
//
// .SECTION See Also
// vtkGenericAdaptorCell vtkDataSet

#ifndef __vtkEdgeFaceDataSet_h
#define __vtkEdgeFaceDataSet_h

#include "vtkGenericDataSet.h"

class vtkPoints;
class vtkCellArray;
class vtkEdgeFaceLocator;
class vtkIdTypeArray;
class vtkUnsignedCharArray;

class VTK_EXPORT vtkEdgeFaceDataSet : public vtkGenericDataSet
{
public:
  // Description:
  // Standard VTK type and print macros.
  static vtkEdgeFaceDataSet* New();
  vtkTypeRevisionMacro(vtkEdgeFaceDataSet,vtkGenericDataSet);
  void PrintSelf(ostream& os, vtkIndent indent);

  // Description:
  // Return the number of points composing the dataset. See NewPointIterator()
  // for more details.
  // \post positive_result: result>=0
  virtual vtkIdType GetNumberOfPoints();

  // Description:
  // Return the number of cells that explicitly define the dataset. See
  // NewCellIterator() for more details.
  // \pre valid_dim_range: (dim>=-1) && (dim<=3)
  // \post positive_result: result>=0
  virtual vtkIdType GetNumberOfCells( int dim = -1 );

  // Description:
  // Return -1 if the dataset is explicitly defined by cells of varying
  // dimensions or if there are no cells. If the dataset is explicitly
  // defined by cells of a unique dimension, return this dimension.
  // \post valid_range: (result>=-1) && (result<=3)
  virtual int GetCellDimension();

  // Description:
  // Return an iterator to traverse cells of dimension `dim' (or all
  // dimensions if -1) that explicitly define the dataset. For instance, it
  // will return only tetrahedra if the mesh is defined by tetrahedra. If the
  // mesh is composed of two parts, one with tetrahedra and another part with
  // triangles, it will return both, but will not return the boundary edges
  // and vertices of these cells. The user is responsible for deleting the
  // iterator.
  // \pre valid_dim_range: (dim>=-1) && (dim<=3)
  // \post result_exists: result!=0
  virtual vtkGenericCellIterator* NewCellIterator( int dim = -1 );

  // Description:
  // Return an iterator to traverse cell boundaries of dimension `dim' (or
  // all dimensions if -1) of the dataset.  If `exteriorOnly' is true, only
  // the exterior cell boundaries of the dataset will be returned, otherwise
  // it will return exterior and interior cell boundaries. The user is
  // responsible for deleting the iterator.
  // \pre valid_dim_range: (dim>=-1) && (dim<=2)
  // \post result_exists: result!=0
  virtual vtkGenericCellIterator* NewBoundaryIterator( int dim=-1, int exteriorOnly = 0 );

  // Description:
  // Return an iterator to traverse the points composing the dataset; they
  // can be points that define a cell (corner points) or isolated points.
  // The user is responsible for deleting the iterator.
  // \post result_exists: result!=0
  virtual vtkGenericPointIterator* NewPointIterator();

  // Description:
  // Locate the closest cell to position `x' (global coordinates) with
  // respect to a tolerance squared `tol2' and an initial guess `cell' (if
  // valid). The result consists in the `cell', the `subId' of the sub-cell
  // (0 if primary cell), the parametric coordinates `pcoord' of the
  // position. It returns whether the position is inside the cell or
  // not (boolean). Tolerance is used to control how close the point is to be
  // considered "in" the cell.
  // THIS METHOD IS NOT THREAD SAFE.
  // \pre not_empty: GetNumberOfCells()>0
  // \pre cell_exists: cell!=0
  // \pre positive_tolerance: tol2>0
  virtual int FindCell(
    double x[3], vtkGenericCellIterator*& cell, double tol2,
    int& subId, double pcoords[3] );

  // Description:
  // Locate the closest point `p' to position `x' (global coordinates).
  // \pre not_empty: GetNumberOfPoints()>0
  // \pre p_exists: p!=0
  virtual void FindPoint( double x[3], vtkGenericPointIterator* p );

  // Description:
  // Compute the geometry bounding box.
  virtual void ComputeBounds();

  // Description:
  // Actual size of the data in kilobytes; only valid after the pipeline has
  // updated. It is guaranteed to be greater than or equal to the memory
  // required to represent the data.
  virtual unsigned long GetActualMemorySize();

  // Description:
  // Return the type of data object.
  int GetDataObjectType();

  // Description:
  // Estimated size needed after tessellation (or special operation)
  virtual vtkIdType GetEstimatedSize();

  //BTX
  // Description:
  // Retrieve an instance of this class from an information object.
  static vtkEdgeFaceDataSet* GetData( vtkInformation* info );
  static vtkEdgeFaceDataSet* GetData( vtkInformationVector* v, int i = 0 );
  //ETX

  // Description:
  // Provide access to the connectivity array of the hexahedra in the mesh.
  vtkGetObjectMacro(Connectivity,vtkCellArray);
  virtual void SetConnectivity( vtkCellArray* );

  // Description:
  // Provide access to the topological corner vertices defining the mesh.
  vtkGetObjectMacro(Points,vtkPoints);
  virtual void SetPoints( vtkPoints* );

  // Description:
  // Provide access to a point locator.
  vtkGetObjectMacro(Locator,vtkEdgeFaceLocator);
  virtual void SetLocator( vtkEdgeFaceLocator* );

  // Description:
  // A convenience method to retrieve a mesh corner point's geometric coordinates.
  void GetPoint( vtkIdType pt, double* x );
  double* GetPoint( vtkIdType pt );

  // Description:
  // Compute cell lookup, neighbor ids, and neighbor faces from connectivity.
  // See the GetNeighborIds and GetNeighborFaces methods for more information.
  virtual void ComputeNeighbors();

  // Description:
  // Get/set the array of neighbors for each cell.
  //
  // All cells are hexahedra and have 6 neighbors.
  // The NeighborIds array has 6 components and as many tuples as there are cells.
  // The entry in component 0 is the Id of the cell neighboring on the -X face.
  // Component 1 is +X, followed by -Y, +Y, -Z, and +Z.
  // Where a cell has no neighbor on a given face, the Id will be set to -1.
  // This array, along with the NeighborFaces array, may be used to traverse the mesh spatially.
  virtual void SetNeighborIds( vtkIdTypeArray* );
  vtkGetObjectMacro(NeighborIds,vtkIdTypeArray);

  // Description:
  // Get/set the array specifying which face of a neighbor corresponds to this cell.
  //
  // The NeighborIds array uses component number to identify neighbors on a given face.
  // This array indicates which component in the neighbor's NeighborIds entry contains a given cell.
  // For instance, a 2-hex mesh with point connectivity [ 0, 1, 2, 3, 4, 5, 6, 7 ]
  // and [ 4, 7, 11, 8, 5, 6, 10, 9 ] would have NeighborIds tuple 0 set to [ -1, -1, -1, -1, -1, 1 ]
  // NeighborIds tuple 1 set to [ -1, -1, 0, -1, -1, -1 ] with NeighborFaces tuple 0
  // equal to [ 255, 255, 255, 255, 255, 2 ] and NeighborFaces tuple 1 equal to [ 255, 255, 5, 255, 255, 255 ].
  virtual void SetNeighborFaces( vtkUnsignedCharArray* );
  vtkGetObjectMacro(NeighborFaces,vtkUnsignedCharArray);

  // Description:
  // The CellLookup array has the same number of tuples as there are points and is used to
  // determine the ID of a cell referencing a given point.
  // If no cell references a given point, its entry in CellLookup will be -1.
  virtual void SetCellLookup( vtkIdTypeArray* );
  vtkGetObjectMacro(CellLookup,vtkIdTypeArray);

protected:
  vtkEdgeFaceDataSet();
  virtual ~vtkEdgeFaceDataSet();

  vtkPoints* Points;
  vtkCellArray* Connectivity;
  vtkEdgeFaceLocator* Locator;
  vtkIdTypeArray* NeighborIds;
  vtkUnsignedCharArray* NeighborFaces;
  vtkIdTypeArray* CellLookup;

private:
  vtkEdgeFaceDataSet( const vtkEdgeFaceDataSet& ); // Not implemented.
  void operator = ( const vtkEdgeFaceDataSet& ); // Not implemented.
};

#endif
