// .NAME vtkEdgeFacePointIterator - iterator used to traverse points
// .SECTION Description
// This class (and subclasses) are used to iterate over points. Use it
// only in conjunction with vtkEdgeFaceDataSet (i.e., the adaptor framework).
//
// Typical use is:
// <pre>
// vtkEdgeFaceDataSet *dataset;
// vtkEdgeFacePointIterator *it = dataset->NewPointIterator();
// for (it->Begin(); !it->IsAtEnd(); it->Next());
//   {
//   x=it->GetPosition();
//   }
// </pre>


#ifndef __vtkEdgeFacePointIterator_h
#define __vtkEdgeFacePointIterator_h

#include "vtkGenericPointIterator.h"

class vtkEdgeFaceDataSet;

class VTK_FILTERING_EXPORT vtkEdgeFacePointIterator : public vtkGenericPointIterator
{
public:
  // Description:
  // Standard VTK construction and type macros.
  static vtkEdgeFacePointIterator* New();
  vtkTypeRevisionMacro(vtkEdgeFacePointIterator,vtkGenericPointIterator);
  void PrintSelf( ostream& os, vtkIndent indent );

  // Description:
  // Move iterator to first position if any (loop initialization).
  virtual void Begin();

  // Description:
  // Is the iterator at the end of traversal?
  virtual int IsAtEnd();

  // Description:
  // Move the iterator to the next position in the list.
  // \pre not_off: !IsAtEnd()
  virtual void Next();
 
  // Description:
  // Get the coordinates of the point at the current iterator position.
  // \pre not_off: !IsAtEnd()
  // \post result_exists: result!=0
  virtual double* GetPosition();

  // Description:
  // Get the coordinates of the point at the current iterator position.
  // \pre not_off: !IsAtEnd()
  // \pre x_exists: x!=0
  virtual void GetPosition( double x[3] );
  
  // Description:
  // Return the unique identifier for the point, could be non-contiguous.
  // \pre not_off: !IsAtEnd()
  virtual vtkIdType GetId();

  //BTX
  /**!\brief The ways in which iteration can progress. */
  enum IterationMode
    {
    MESH,    //!< Iterate over every point in the mesh
    CELL,    //!< Iterate over every point in a single cell
    BOUNDARY //!< Iterate over every point on a d-dimensional boundary of a single cell.
    };
  // Description:
  // Provide a way to iterate over points of a cell or one of its boundaries.
  vtkSetMacro(Mode,IterationMode);

  // Description:
  // Return the mode, describing how the iterator's Next() method will progress.
  vtkGetMacro(Mode,IterationMode)
  //ETX
 
  // Description:
  // When the Mode is set to CELL or BOUNDARY, this contains the cell of reference.
  vtkSetMacro(CellId,vtkIdType);
  vtkGetMacro(CellId,vtkIdType);

  // Description:
  // When the Mode is set to BOUNDARY, this indicates the dimension of boundary to traverse.
  // It should be 1 (to get the points on an edge)
  // or 2 (to get the points on the face of a volumetric cell).
  // Any other values are invalid.
  vtkSetMacro(Dimension,int);
  vtkGetMacro(Dimension,int);

  // Description:
  // When the Mode is set to BOUNDARY, this indicates which boundary of the specified Dimension
  // should be traversed.
  // Since we only handle hexahedra and quadrilaterals, valid values are
  // - {0,1,2,3,4,5} when the cell is a hex and Dimension is 2;
  // - {0,1,...,10,11} when the cell is a hex and Dimension is 1;
  // - {0,1,2,3} when the cell is a quad and Dimension is 1.
  vtkSetMacro(Boundary,int);
  vtkGetMacro(Boundary,int);

  // Description:
  // When the mode is set to CELL or BOUNDARY, this is used to store which point on the cell
  // or boundary facet the iterator is currently at.
  vtkSetMacro(Index,int);
  vtkGetMacro(Index,int);

protected:
  // Description:
  // Destructor.
  vtkEdgeFacePointIterator();
  virtual ~vtkEdgeFacePointIterator();
  
  virtual void SetDataSet( vtkEdgeFaceDataSet* dataSet );
  vtkEdgeFaceDataSet* GetDataSet() { return this->DataSet; }

  // Description:
  // Set the point currently being visited.
  // This is only for use by vtkEdgeFaceDataSet
  vtkSetMacro(Id,vtkIdType);
 
  //BTX 
  friend class vtkEdgeFaceDataSet;
  vtkEdgeFaceDataSet* DataSet;
  vtkIdType Id;
  IterationMode Mode;
  vtkIdType CellId; 
  int Dimension;
  int Boundary;
  int Index;
  //ETX

private:
  vtkEdgeFacePointIterator( const vtkEdgeFacePointIterator& ); // Not implemented.
  void operator = ( const vtkEdgeFacePointIterator& ); // Not implemented.
};

#endif
