#include "vtkEdgeFaceReader.h"

#include "vtkEdgeFaceDataSet.h"
#include "vtkEdgeFaceAttribute.h"

#include "vtkCellArray.h"
#include "vtkDoubleArray.h"
#include "vtkGenericAttributeCollection.h"
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkObjectFactory.h"
#include "vtkPoints.h"

#include <vtkstd/string>
#include <vtkstd/vector>
//#include <vtkstd/ifstream>

#include "vtksys/SystemTools.hxx"

vtkStandardNewMacro(vtkEdgeFaceReader);
vtkCxxRevisionMacro(vtkEdgeFaceReader,"$Revision:$");

class vtkEdgeFaceReader::Private
{
public:
  vtkIdType NumberOfPoints;
  vtkIdType NumberOfEdges;
  vtkIdType NumberOfFaces;
  vtkIdType NumberOfHexes;
  vtkstd::vector<vtkstd::string> EdgeFieldNames;
  vtkstd::vector<vtkstd::string> FaceFieldNames;
  off_t DataOffset;
  long int ModificationTime;
};

vtkEdgeFaceReader::vtkEdgeFaceReader()
{
  this->Data = new vtkEdgeFaceReader::Private;
  this->FileName = 0;
  this->SetNumberOfInputPorts( 0 );
}

vtkEdgeFaceReader::~vtkEdgeFaceReader()
{
  delete this->Data;
  this->SetFileName( 0 );
}

void vtkEdgeFaceReader::PrintSelf( ostream& os, vtkIndent indent )
{
  this->Superclass::PrintSelf( os, indent );
  if ( this->FileName )
    {
    os << "FileName: \"" << this->FileName << "\"\n";
    }
  else
    {
    os << "FileName: (null)\n";
    }
  os << "Data: " << this->Data << "\n";
}

int vtkEdgeFaceReader::RequestDataObject(
  vtkInformation* vtkNotUsed(request), vtkInformationVector** vtkNotUsed(inInfo), vtkInformationVector* outInfo )
{
  if ( ! outInfo )
    {
    return 0;
    }
  vtkInformation* info = outInfo->GetInformationObject( 0 );
  if ( ! info )
    {
    return 0;
    }

  vtkEdgeFaceDataSet* output = vtkEdgeFaceDataSet::SafeDownCast( info->Get( vtkDataObject::DATA_OBJECT() ) );
  if ( ! output )
    {
    output = vtkEdgeFaceDataSet::New();
    output->SetPipelineInformation( info );
    output->Delete();
    this->GetOutputPortInformation( 0 )->Set( vtkDataObject::DATA_EXTENT_TYPE(), output->GetExtentType() );
    }

  return 1;
}

int vtkEdgeFaceReader::RequestInformation(
  vtkInformation* vtkNotUsed(request), vtkInformationVector** vtkNotUsed(inInfo), vtkInformationVector* outInfo )
{
  if ( ! outInfo )
    {
    return 0;
    }
  vtkInformation* info = outInfo->GetInformationObject( 0 );
  if ( ! info )
    {
    return 0;
    }

  vtkEdgeFaceDataSet* output = vtkEdgeFaceDataSet::SafeDownCast( info->Get( vtkDataObject::DATA_OBJECT() ) );
  if ( ! output )
    {
    return 0;
    }

  return this->ReadMetadata( this->FileName, this->Data, output );
}

int vtkEdgeFaceReader::RequestData(
  vtkInformation* vtkNotUsed(request), vtkInformationVector** vtkNotUsed(inInfo), vtkInformationVector* outInfo )
{
  if ( ! outInfo )
    {
    return 0;
    }
  vtkInformation* info = outInfo->GetInformationObject( 0 );
  if ( ! info )
    {
    return 0;
    }

  vtkEdgeFaceDataSet* output = vtkEdgeFaceDataSet::SafeDownCast( info->Get( vtkDataObject::DATA_OBJECT() ) );
  if ( ! output )
    {
    return 0;
    }

  return this->ReadData( this->FileName, this->Data, output );
}

int vtkEdgeFaceReader::ReadMetadata( const char* fname, vtkEdgeFaceReader::Private* meta, vtkEdgeFaceDataSet* ef )
{
  vtkstd::ifstream data( fname );
  if ( data.fail() )
    {
    return 0;
    }
  meta->ModificationTime = vtksys::SystemTools::ModifiedTime( fname );

  vtkIdType num;
  vtkstd::string name;

  data >> meta->NumberOfPoints;
  //ef->GetPoints()->SetNumberOfPoints( meta->NumberOfPoints );

  data >> meta->NumberOfEdges;
  data >> meta->NumberOfFaces;

  data >> meta->NumberOfHexes;
  ef->GetConnectivity()->SetNumberOfCells( meta->NumberOfHexes );

  data >> num; // number of edge fields
  for ( vtkIdType i = 0; i < num; ++ i )
    {
    data >> name;
    meta->EdgeFieldNames.push_back( name );
    }

  data >> num; // number of face fields
  for ( vtkIdType i = 0; i < num; ++ i )
    {
    data >> name;
    meta->FaceFieldNames.push_back( name );
    }
  meta->DataOffset = data.tellg();

  return 1;
}

int vtkEdgeFaceReader::ReadData( const char* fname, vtkEdgeFaceReader::Private* meta, vtkEdgeFaceDataSet* ef )
{
  long int fmtime = vtksys::SystemTools::ModifiedTime( fname );
  if ( fmtime > meta->ModificationTime )
    { // Do we need to re-read header info?
    if ( ! this->ReadMetadata( fname, meta, ef ) )
      {
      return 0;
      }
    }

  vtkstd::ifstream data( fname );
  if ( data.fail() )
    {
    return 0;
    }
  data.seekg( meta->DataOffset, vtkstd::ios_base::beg );
  if ( data.fail() )
    {
    return 0;
    }

  double x, y, z;
  vtkIdType i;
  vtkIdType entries[8 + 12 + 6];

  vtkPoints* pts = ef->GetPoints();
  pts->SetNumberOfPoints( meta->NumberOfPoints );
  for ( i = 0; i < meta->NumberOfPoints; ++ i )
    {
    data >> x;
    data >> y;
    data >> z;
    pts->SetPoint( i, x, y, z );
    }
  pts->Modified();

  vtkCellArray* conn = ef->GetConnectivity();
  conn->Reset();
  for ( i = 0; i < meta->NumberOfHexes; ++ i )
    {
    for ( int j = 0; j < ( 8 + 6 + 12 ); ++ j )
      {
      data >> entries[j];
      }
    conn->InsertNextCell( 8 + 6 + 12, entries );
    }
  ef->ComputeNeighbors();

  vtkstd::vector<vtkstd::string>::iterator efit;
  for ( efit = meta->EdgeFieldNames.begin(); efit != meta->EdgeFieldNames.end(); ++ efit )
    {
    vtkEdgeFaceAttribute* att = vtkEdgeFaceAttribute::New();
    att->SetDeRhamBasis( vtkEdgeFaceAttribute::EDGE );
    vtkDoubleArray* arr = att->GetArray();
    arr->SetName( efit->c_str() );
    arr->SetNumberOfTuples( meta->NumberOfEdges );
    for ( i = 0; i < meta->NumberOfEdges; ++ i )
      {
      data >> x;
      arr->SetValue( i, x );
      }
    ef->GetAttributes()->InsertNextAttribute( att );
    att->Delete();
    }

  vtkstd::vector<vtkstd::string>::iterator ffit;
  for ( ffit = meta->FaceFieldNames.begin(); ffit != meta->FaceFieldNames.end(); ++ ffit )
    {
    vtkEdgeFaceAttribute* att = vtkEdgeFaceAttribute::New();
    att->SetDeRhamBasis( vtkEdgeFaceAttribute::FACE );
    vtkDoubleArray* arr = att->GetArray();
    arr->SetName( ffit->c_str() );
    arr->SetNumberOfTuples( meta->NumberOfFaces );
    for ( i = 0; i < meta->NumberOfFaces; ++ i )
      {
      data >> x;
      arr->SetValue( i, x );
      }
    ef->GetAttributes()->InsertNextAttribute( att );
    att->Delete();
    }
  return 1;
}

