#ifndef __vtkEdgeFaceReader_h
#define __vtkEdgeFaceReader_h
// .NAME vtkEdgeFaceReader - Read in finite element data for edge- and face-elements.
// .SECTION Description
// This reader is very simple.
// It only handles hexahedra.
// Its RequestInformation() member reads in
// - the number of points,
// - the number of edges,
// - the number of faces,
// - the number of hexahedra,
// - the number and names of edge fields, and
// - the number and names of face fields.
// The RequestData() member reads in
// - the points,
// - the connectivity array,
// - all edge field data,
// - all face field data.
// You may not select a subset of fields to be read.
// This is example code.

#include "vtkGenericDataSetAlgorithm.h"

class vtkEdgeFaceDataSet;

class VTK_EXPORT vtkEdgeFaceReader : public vtkGenericDataSetAlgorithm
{
public:
  static vtkEdgeFaceReader* New();
  virtual void PrintSelf( ostream& os, vtkIndent indent );
  vtkTypeRevisionMacro(vtkEdgeFaceReader,vtkGenericDataSetAlgorithm);

  // Description:
  // Set/get the filename containing an edge/face dataset.
  vtkGetStringMacro(FileName);
  vtkSetStringMacro(FileName);

protected:
  vtkEdgeFaceReader();
  virtual ~vtkEdgeFaceReader();

  // Description:
  // Create an output dataset of the correct type (i.e., create a vtkEdgeFaceDataSet) and point \a outInfo to it.
  // @param request A dictionary containing information that the pipeline should use to prepare \a outInfo.
  // @param inInfo An array of arrays of dictionaries containing information for each of this filter's inputs.
  //               Not used since this is a reader and has no input ports.
  // @param outInfo An array of dictionaries containing information on each output port of the reader.
  virtual int RequestDataObject( vtkInformation* request, vtkInformationVector** inInfo, vtkInformationVector* outInfo );

  // Description:
  // Read metadata from the front of the file.
  // This is invoked so that downstream filters can prepare to handle the output.
  // @param request A dictionary containing information that the pipeline should use to prepare \a outInfo.
  // @param inInfo An array of arrays of dictionaries containing information for each of this filter's inputs.
  //               Not used since this is a reader and has no input ports.
  // @param outInfo An array of dictionaries containing information on each output port of the reader.
  virtual int RequestInformation( vtkInformation* request, vtkInformationVector** inInfo, vtkInformationVector* outInfo );

  // Description:
  // Read the actual data contained in the file and populate a dataset with the information requested.
  // @param request A dictionary containing information that the pipeline should use to prepare \a outInfo.
  // @param inInfo An array of arrays of dictionaries containing information for each of this filter's inputs.
  //               Not used since this is a reader and has no input ports.
  // @param outInfo An array of dictionaries containing information on each output port of the reader.
  virtual int RequestData( vtkInformation* request, vtkInformationVector** inInfo, vtkInformationVector* outInfo );

  //BTX
  class Private;
  //ETX

  //BTX
  // Description:
  // Read metadata from the file header and save it in the private data object:
  int ReadMetadata( const char* fname, Private* meta, vtkEdgeFaceDataSet* ef );
  //ETX

  //BTX
  // Description:
  // Read data and populate a mesh. This is called from within RequestData.
  int ReadData( const char* fname, Private* meta, vtkEdgeFaceDataSet* ef );
  //ETX

  //BTX
  class Private;
  Private* Data;
  char* FileName;
  //ETX

private:
  vtkEdgeFaceReader( const vtkEdgeFaceReader& ); // Not implemented.
  void operator = ( const vtkEdgeFaceReader& ); // Not implemented.
};

#endif // __vtkEdgeFaceReader_h
